"""Percentile calculation functions."""

import numpy as np
import pandas as pd
from hyswap.utils import filter_data_by_time
from hyswap.utils import calculate_metadata
from hyswap.utils import define_year_doy_columns


def calculate_fixed_percentile_thresholds(
        data,
        percentiles=np.array((0, 5, 10, 25, 75, 90, 95, 100)),
        method='weibull',
        **kwargs):
    """Calculate fixed percentile thresholds using historic data.

    Parameters
    ----------
    data : array_like
        1D array of data from which to calculate percentile thresholds.

    percentiles : array_like, optional
        Percentiles to calculate. Default is (0, 5, 10, 25, 75, 90, 95, 100).

    method : str, optional
        Method to use to calculate percentiles. Default is 'weibull'.

    **kwargs : dict, optional
        Additional keyword arguments to pass to `numpy.percentile`.

    Returns
    -------
    percentiles : array_like
        Percentiles of the data.

    Examples
    --------
    Calculate default percentile thresholds from some synthetic data.

    .. doctest::

        >>> data = np.arange(101)
        >>> results = percentiles.calculate_fixed_percentile_thresholds(
        ...     data, method='linear')
        >>> results
        array([  0.,   5.,  10.,  25.,  75.,  90.,  95., 100.])

    Calculate a different set of thresholds from some synthetic data.

    .. doctest::

        >>> data = np.arange(101)
        >>> results = percentiles.calculate_fixed_percentile_thresholds(
        ...     data, percentiles=np.array((0, 10, 50, 90, 100)))
        >>> results
        array([  0. ,   9.2,  50. ,  90.8, 100. ])
    """
    return np.percentile(data, percentiles, method=method, **kwargs)


def calculate_variable_percentile_thresholds_by_day(
        df,
        data_column_name,
        percentiles=[0, 5, 10, 25, 75, 90, 95, 100],
        method='weibull',
        date_column_name=None,
        year_type='calendar',
        min_years=10,
        **kwargs):
    """Calculate variable percentile thresholds of data by day of year.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame containing data to calculate daily percentile thresholds for.

    data_column_name : str
        Name of column containing data to analyze.

    percentiles : array_like, optional
        Percentile thresholds to calculate, default is
        [0, 5, 10, 25, 75, 90, 95, 100].

    method : str, optional
        Method to use to calculate percentiles. Default is 'weibull'.

    date_column_name : str, optional
        Name of column containing date information. If None, the index of
        `df` is used.

    year_type : str, optional
        The type of year to use. Must be one of 'calendar', 'water', or
        'climate'. Default is 'calendar' which starts the year on January 1
        and ends on December 31. 'water' starts the year on October 1 and
        ends on September 30 of the following year which is the "water year".
        For example, October 1, 2010 to September 30, 2011 is "water year
        2011". 'climate' years begin on April 1 and end on March 31 of the
        following year, they are numbered by the ending year. For example,
        April 1, 2010 to March 31, 2011 is "climate year 2011".

    min_years : int, optional
        Minimum number of years of data required to calculate percentile
        thresholds for a given day of year. Default is 10.

    **kwargs : dict, optional
        Additional keyword arguments to pass to `numpy.percentile`.

    Returns
    -------
    percentiles : pandas.DataFrame
        DataFrame containing threshold percentiles of data by day of year.

    Examples
    --------
    Calculate default thresholds by day of year from some real data in
    preparation for plotting.

    .. doctest::
        :skipif: True  # dataretrieval functions break CI pipeline

        >>> df, _ = dataretrieval.nwis.get_dv(
        ...     "03586500", parameterCd="00060",
        ...     start="1776-01-01", end="2022-12-31")
        >>> results = percentiles.calculate_variable_percentile_thresholds_by_day(  # noqa: E501
        ...     df, "00060_Mean")
        >>> len(results.index)  # 366 days in a leap year
        366
        >>> len(results.columns)  # 8 default percentiles
        8
    """
    # define year and day of year columns and convert date column to datetime
    # if necessary
    df = define_year_doy_columns(df, date_column_name=date_column_name,
                                 year_type=year_type, clip_leap_day=True)
    # based on date, get min and max day of year available
    min_day = np.nanmax((1, df.index.dayofyear.min()))
    max_day = np.nanmin((366, df.index.dayofyear.max() + 1))
    # make temporal index
    t_idx = np.arange(min_day, max_day)
    # initialize a DataFrame to hold percentiles by day of year
    percentiles_by_day = pd.DataFrame(index=t_idx, columns=percentiles)
    # loop through days of year available
    for doy in range(min_day, max_day):
        # get historical data for the day of year
        data = filter_data_by_time(df, doy, data_column_name)
        # could insert other functions here to check or modify data
        # as needed or based on any other criteria
        meta = calculate_metadata(data)

        # only calculate data if there are at least min_years of data
        if meta['n_years'] >= min_years:
            # calculate percentiles for the day of year and add to DataFrame
            percentiles_by_day.loc[t_idx == doy, :] = \
                calculate_fixed_percentile_thresholds(
                    data, percentiles=percentiles, method=method, **kwargs)
        else:
            # if there are not at least 10 years of data,
            # set percentiles to NaN
            percentiles_by_day.loc[t_idx == doy, :] = np.nan

    # replace index with multi-index of doy_index and month-day values
    # month_day values
    month_day = pd.to_datetime(
        percentiles_by_day.index, format='%j').strftime('%m-%d')
    # doy_index values
    doy_index = percentiles_by_day.index.values
    if year_type == 'water':
        doy_index = doy_index - 273
        doy_index[doy_index < 1] += 365
    elif year_type == 'climate':
        doy_index = doy_index - 90
        doy_index[doy_index < 1] += 365
    percentiles_by_day.index = pd.MultiIndex.from_arrays(
        [doy_index, month_day], names=['doy', 'month-day'])

    # sort percentiles by index
    percentiles_by_day.sort_index(inplace=True)

    # return percentiles by day of year
    return percentiles_by_day



"""Unit tests for the rasterhydrograph module."""
import pytest
import numpy as np
import pandas as pd
from hyswap import rasterhydrograph


def test_set_data_type():
    """Test the private function _set_data_type."""
    assert rasterhydrograph._set_data_type('daily') == 'D'
    assert rasterhydrograph._set_data_type('7-day') == '7D'
    assert rasterhydrograph._set_data_type('14-day') == '14D'
    assert rasterhydrograph._set_data_type('28-day') == '28D'


def test_calculate_date_range():
    """Test the private function _calculate_date_range."""
    # define a data frame to use for testing
    dummy_dates = pd.date_range('2018-06-01', '2022-01-31')
    df = pd.DataFrame(
        {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates)),
         'index_year': dummy_dates.year.tolist(),
         'month': dummy_dates.month.tolist(),
         'day': dummy_dates.day.tolist(),
         'index_doy': dummy_dates.dayofyear.tolist()}
        )
    df.set_index('date', inplace=True)
    # test the function with default values
    date_range = rasterhydrograph._calculate_date_range(df, 'calendar',
                                                        None, None)
    assert date_range[0].year == 2018
    assert date_range[0].month == 1
    assert date_range[0].day == 1
    assert date_range[-1].year == 2022
    assert date_range[-1].month == 12
    assert date_range[-1].day == 31
    # test the function specifying start and end years
    date_range = rasterhydrograph._calculate_date_range(df, 'calendar',
                                                        2019, 2020)
    assert date_range[0].year == 2019
    assert date_range[0].month == 1
    assert date_range[0].day == 1
    assert date_range[-1].year == 2020
    assert date_range[-1].month == 12
    assert date_range[-1].day == 31


def test_check_inputs():
    """Test the private function _check_inputs."""
    # make a df
    dummy_dates = pd.date_range('2018-06-01', '2022-01-31')
    df = pd.DataFrame(
        {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates))}
        )
    df_date = df.set_index('date')
    # test the data frame input
    with pytest.raises(TypeError):
        rasterhydrograph._check_inputs(None, None, None, None,
                                       None, None, None)
    # test the data column name input
    with pytest.raises(TypeError):
        rasterhydrograph._check_inputs(pd.DataFrame(), None, None, 'daily',
                                       'calendar', None, None)
    with pytest.raises(TypeError):
        rasterhydrograph._check_inputs(pd.DataFrame(), 1, None, 'daily',
                                       'calendar', None, None)
    with pytest.raises(TypeError):
        rasterhydrograph._check_inputs(pd.DataFrame(), 1.0, None, 'daily',
                                       'calendar', None, None)

    # test the date column name input
    with pytest.raises(TypeError):
        rasterhydrograph._check_inputs(pd.DataFrame(), 'data', {"a": 1},
                                       'daily', 'calendar', None, None)
    with pytest.raises(TypeError):
        rasterhydrograph._check_inputs(pd.DataFrame(), 'data', 1, 'daily',
                                       'calendar', None, None)
    with pytest.raises(TypeError):
        rasterhydrograph._check_inputs(pd.DataFrame(), 'data', 1.0, 'daily',
                                       'calendar', None, None)

    # test the data type input
    with pytest.raises(TypeError):
        rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None, 1,
                                       None, None, None)
    with pytest.raises(TypeError):
        rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None, 1.0,
                                       None, None, None)
    with pytest.raises(TypeError):
        rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None, None,
                                       None, None, None)
    with pytest.raises(ValueError):
        rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None, 'string',
                                       None, None, None)

    # test the year type input
    with pytest.raises(ValueError):
        rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None, 'daily',
                                       1, None, None)
    with pytest.raises(ValueError):
        rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None, 'daily',
                                       1.0, None, None)
    with pytest.raises(ValueError):
        rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None, 'daily',
                                       None, None, None)
    with pytest.raises(ValueError):
        rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None, 'daily',
                                       'string', None, None)

    # test the begin year input
    with pytest.raises(TypeError):
        rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None, 'daily',
                                       'calendar', 1.0, None)
    with pytest.raises(TypeError):
        rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None, 'daily',
                                       'calendar', 'string', None)
    with pytest.raises(ValueError):
        rasterhydrograph._check_inputs(df, 'data', 'date', 'daily',
                                       'calendar', 2000, None)
    with pytest.raises(ValueError):
        rasterhydrograph._check_inputs(df_date, 'data', None, 'daily',
                                       'calendar', 2000, None)

    # test the end year input
    with pytest.raises(TypeError):
        rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None, 'daily',
                                       'calendar', None, 1.0)
    with pytest.raises(TypeError):
        rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None, 'daily',
                                       'calendar', None, 'string')
    with pytest.raises(ValueError):
        rasterhydrograph._check_inputs(df, 'data', 'date', 'daily',
                                       'calendar', None, 3000)
    with pytest.raises(ValueError):
        rasterhydrograph._check_inputs(df_date, 'data', None, 'daily',
                                       'calendar', None, 3000)


def test_format_data():
    """Test the public function format_data."""
    # set up some data for the test
    dummy_dates = pd.date_range('2018-06-01', '2022-01-31')
    df = pd.DataFrame(
        {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates))}
    )
    df_date = df.set_index('date')
    # test the function with a dataframe and a date column
    df_out = rasterhydrograph.format_data(df, 'value', 'date')
    assert len(df_out.index) == 5
    assert len(df_out.columns) == 365
    # assert day 1 of 2018 has no data, is NaN
    assert np.isnan(df_out.loc[2018][0])
    # assert day 350 of 2022 has no data, is NaN
    assert np.isnan(df_out.loc[2022][349])
    # test the function with a dataframe and a date index
    df_out = rasterhydrograph.format_data(df_date, 'value')
    assert len(df_out.index) == 5
    assert len(df_out.columns) == 365
    # assert day 1 of 2018 has no data, is NaN
    assert np.isnan(df_out.loc[2018][0])
    # assert day 350 of 2022 has no data, is NaN
    assert np.isnan(df_out.loc[2022][349])
    # test the function with a dataframe and a date index and beginning year
    df_out = rasterhydrograph.format_data(df_date, 'value', begin_year=2019)
    assert len(df_out.index) == 4
    assert len(df_out.columns) == 365
    # assert day 1 of 2019 has data, is not NaN
    assert ~np.isnan(df_out.loc[2019][0])
    # assert day 350 of 2022 has no data, is NaN
    assert np.isnan(df_out.loc[2022][349])
    # test the function with a dataframe and a date index and ending year
    df_out = rasterhydrograph.format_data(df_date, 'value', end_year=2021)
    assert len(df_out.index) == 4
    assert len(df_out.columns) == 365
    # assert day 1 of 2018 has no data, is NaN
    assert np.isnan(df_out.loc[2018][0])
    # assert day 350 of 2021 has data, is not NaN
    assert ~np.isnan(df_out.loc[2021][349])
    # test the function with a dataframe and a date index and a different
    # data averaging scheme
    df_7out = rasterhydrograph.format_data(df_date, 'value',
                                           data_type='7-day')
    assert len(df_7out.index) == 5
    assert len(df_7out.columns) == 365
    # assert day 1 of 2018 has no data, is NaN
    assert np.isnan(df_7out.loc[2018][0])
    # assert day 350 of 2022 has no data, is NaN
    assert np.isnan(df_7out.loc[2022][349])
    # check that there are non-NaN values in the data frame
    assert ~np.isnan(df_7out.values).all()
    # test the function with a dataframe and a date index and a water year
    df_out_water = rasterhydrograph.format_data(df_date, 'value',
                                                year_type='water')
    assert len(df_out_water.index) == 5
    assert len(df_out_water.columns) == 365
    # check that there are non-NaN values in the data frame
    assert ~np.isnan(df_out_water.values).all()
    # check that day 1 of year 2019 is 10/1/2018
    assert df_out_water.loc[2019][0] == df_out.loc[2018][273]
    # check that the last day of year 2020 is 9/30/2020
    assert df_out_water.loc[2020][364] == df_out.loc[2020][272]
    # test the function with a dataframe and a date index and a different
    # data averaging scheme and a water year
    df_7out_water = rasterhydrograph.format_data(df_date, 'value',
                                                 data_type='7-day',
                                                 year_type='water')
    assert len(df_7out_water.index) == 5
    assert len(df_7out_water.columns) == 365
    # check that there are non-NaN values in the data frame
    assert ~np.isnan(df_7out_water.values).all()
    # check that day 1 of year 2019 is 10/1/2018
    assert df_7out_water.loc[2019][0] == df_7out.loc[2018][273]
    # check that the last day of year 2020 is 9/30/2020
    assert df_7out_water.loc[2020][364] == df_7out.loc[2020][272]
    # test the function with a dataframe and a date index and in a
    # climate year
    df_out_climate = rasterhydrograph.format_data(df_date, 'value',
                                                  year_type='climate')
    assert len(df_out_climate.index) == 4
    assert len(df_out_climate.columns) == 365
    # check that there are non-NaN values in the data frame
    assert ~np.isnan(df_out_climate.values).all()
    # check that day 1 of year 2020 is 4/1/2019
    assert df_out_climate.loc[2020][0] == df_out.loc[2019][90]
    # check that the last day of year 2021 is 3/31/2021
    assert df_out_climate.loc[2021][364] == df_out.loc[2021][89]
    # test the function with a dataframe and a date index and a different
    # data averaging scheme via kwargs and a climate year
    df_7out_climate = rasterhydrograph.format_data(df_date, 'value',
                                                   data_type='7-day',
                                                   year_type='climate',
                                                   center=True)
    assert len(df_7out_climate.index) == 4
    assert len(df_7out_climate.columns) == 365
    # check that there are non-NaN values in the data frame
    assert ~np.isnan(df_7out_climate.values).all()
    # this is a different averaging method so the values will be different
    # set up some data starting earlier in the year
    dummy_dates = pd.date_range('2018-02-01', '2022-01-31')
    df = pd.DataFrame(
        {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates))}
    )
    df_date = df.set_index('date')
    # test the function with a dataframe and a date index and in a
    # climate year
    df_out_climate = rasterhydrograph.format_data(df_date, 'value',
                                                  year_type='climate')
    assert len(df_out_climate.index) == 5
    assert len(df_out_climate.columns) == 365
    # check that there are non-NaN values in the data frame
    assert ~np.isnan(df_out_climate.values).all()



"""Cumulative calculation functions."""

import numpy as np
import pandas as pd
from hyswap.utils import define_year_doy_columns


def calculate_daily_cumulative_values(df, data_column_name,
                                      date_column_name=None,
                                      year_type='calendar'):
    """Calculate daily cumulative values.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame containing data to calculate cumulative values.
    data_column_name : str
        Name of column containing data to calculate cumulative values for.
    date_column_name : str, optional
        Name of column containing date information. If None, the index of
        `df` will be used.
    year_type : str, optional
        The type of year to use. Must be one of 'calendar', 'water', or
        'climate'. Default is 'calendar' which starts the year on January 1
        and ends on December 31. 'water' starts the year on October 1 and
        ends on September 30 of the following year which is the "water year".
        For example, October 1, 2010 to September 30, 2011 is "water year
        2011". 'climate' years begin on April 1 and end on March 31 of the
        following year, they are numbered by the ending year. For example,
        April 1, 2010 to March 31, 2011 is "climate year 2011".

    Returns
    -------
    cumulative_values : pandas.DataFrame
        DataFrame containing daily cumulative values for each year in the
        input DataFrame, rows are dates and columns include years, days, and
        cumulative values.

    Examples
    --------
    Calculate daily cumulative values from some synthetic data.

    .. doctest::

        >>> df = pd.DataFrame({
        ...     "date": pd.date_range("2000-01-01", "2000-12-31"),
        ...     "data": np.arange(366)})
        >>> results = cumulative.calculate_daily_cumulative_values(
        ...     df, "data", date_column_name="date")
        >>> results.columns.tolist()
        ['index_year', 'index_doy', 'cumulative']
    """
    # set date index, add day/year columns with function
    df = define_year_doy_columns(df, date_column_name=date_column_name,
                                 year_type=year_type, clip_leap_day=True)
    # get unique years in the data
    years = df['index_year'].unique()
    # make a dataframe to hold cumulative values for each year
    cdf = pd.DataFrame(index=years, columns=np.arange(1, 366))
    # loop through each year and calculate cumulative values
    for year in years:
        # get data for the year
        year_data = df.loc[df['index_year'] == year, data_column_name]
        # year must be complete
        if len(year_data) == 365:
            # calculate cumulative values and assign to cdf
            cdf.loc[cdf.index == year, :len(year_data)] = \
                year_data.cumsum().values
    # reformat the dataframe
    cdf = _tidy_cumulative_dataframe(cdf, year_type)
    return cdf


def _tidy_cumulative_dataframe(cdf, year_type):
    """Tidy a cumulative dataframe.

    Parameters
    ----------
    cdf : pandas.DataFrame
        DataFrame containing cumulative values, rows are years, columns are
        days of year.
    year_type : str
        The type of year to use. Must be one of 'calendar', 'water', or
        'climate'. Default is 'calendar' which starts the year on January 1
        and ends on December 31. 'water' starts the year on October 1 and
        ends on September 30 of the following year which is the "water year".
        For example, October 1, 2010 to September 30, 2011 is "water year
        2011". 'climate' years begin on April 1 and end on March 31 of the
        following year, they are numbered by the ending year. For example,
        April 1, 2010 to March 31, 2011 is "climate year 2011".

    Returns
    -------
    cdf : pandas.DataFrame
        DataFrame containing cumulative values, rows are dates, columns include
        years, day of year (doy), and cumulative values.
    """
    # convert cdf to dataframe organized with full dates on the index
    cdf2 = cdf.stack().reset_index()
    cdf2.columns = ["index_year", "index_doy", "cumulative"]
    # create date column
    if year_type == "calendar":
        cdf2["date"] = pd.to_datetime(
            cdf2["index_year"].astype(str) + "-" +
            cdf2["index_doy"].astype(str),
            format="%Y-%j")
    elif year_type == "water":
        cdf2["date"] = pd.to_datetime(
            cdf2["index_year"].astype(str) + "-" +
            cdf2["index_doy"].astype(str),
            format="%Y-%j") + pd.DateOffset(days=273)
    elif year_type == "climate":
        cdf2["date"] = pd.to_datetime(
            cdf2["index_year"].astype(str) + "-" +
            cdf2["index_doy"].astype(str),
            format="%Y-%j") + pd.DateOffset(days=90)
    # set date to index
    cdf2 = cdf2.set_index("date")
    return cdf2



"""Tests for the percentiles functions."""
import numpy as np
import pandas as pd
import pytest
from hyswap import percentiles


def test_calculate_fixed_percentile_thresholds():
    """Test the calculate_fixed_percentile_thresholds function."""
    # make some data
    data = np.arange(101)
    # test the function
    percentiles_ = percentiles.calculate_fixed_percentile_thresholds(
        data, method='linear')
    assert percentiles_.shape == (8,)
    assert percentiles_ == pytest.approx((0, 5, 10, 25, 75, 90, 95, 100))
    # set some percentile values as opposed to the defaults
    percentiles_ = percentiles.calculate_fixed_percentile_thresholds(
        data, percentiles=np.array((0, 10, 50, 90, 100)))
    assert percentiles_.shape == (5,)
    assert percentiles_ == pytest.approx((0, 9.2, 50, 90.8, 100))
    # pass kwarg through to np.percentile
    percentiles_ = percentiles.calculate_fixed_percentile_thresholds(
        data, method='lower')
    assert percentiles_.shape == (8,)
    assert percentiles_ == pytest.approx((0, 5, 10, 25, 75, 90, 95, 100))


def test_calculate_variable_percentile_thresholds_by_day():
    """Test the calculate_variable_percentile_thresholds_by_day function."""
    # make a dataframe
    df = pd.DataFrame({
        'data': np.arange(101),
        'date': pd.date_range('2019-01-01', '2019-04-11')})
    # test the function
    percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day(
        df, 'data', date_column_name='date')
    assert percentiles_.shape == (101, 8)
    assert percentiles_.columns.tolist() == [0, 5, 10, 25, 75, 90, 95, 100]
    assert percentiles_.index.get_level_values(1).tolist()[0] == '01-01'
    assert percentiles_.index.get_level_values(0).tolist()[0] == 1
    # test the function with no date column and dates in the index
    df = df.set_index('date')
    percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day(
        df, 'data')
    assert percentiles_.shape == (101, 8)
    assert percentiles_.columns.tolist() == [0, 5, 10, 25, 75, 90, 95, 100]
    assert percentiles_.index.get_level_values(1).tolist()[0] == '01-01'
    assert percentiles_.index.get_level_values(0).tolist()[0] == 1
    # test the function with a different set of percentiles
    percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day(
        df, 'data', percentiles=np.array((0, 10, 50, 90, 100)))
    assert percentiles_.shape == (101, 5)
    assert percentiles_.columns.tolist() == [0, 10, 50, 90, 100]
    assert percentiles_.index.get_level_values(1).tolist()[0] == '01-01'
    assert percentiles_.index.get_level_values(0).tolist()[0] == 1
    # all percentiles should be NaN because demo dataset is 1 year only
    assert percentiles_.isna().all().all()
    # test the function with a different year type
    percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day(
        df, 'data', year_type='water')
    assert percentiles_.shape == (101, 8)
    assert percentiles_.columns.tolist() == [0, 5, 10, 25, 75, 90, 95, 100]
    assert percentiles_.index.get_level_values(1).tolist()[0] == '01-01'
    assert percentiles_.index.get_level_values(0).tolist()[0] == 93
    assert percentiles_.index.get_level_values(1).tolist()[-1] == '04-11'
    assert percentiles_.index.get_level_values(0).tolist()[-1] == 193
    # all percentiles should be NaN because demo dataset is 1 year only
    assert percentiles_.isna().all().all()
    # test the function with a different year type and a different set of
    # percentiles
    percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day(
        df, 'data', year_type='climate',
        percentiles=np.array((0, 10, 50, 90, 100)))
    assert percentiles_.shape == (101, 5)
    assert percentiles_.columns.tolist() == [0, 10, 50, 90, 100]
    assert percentiles_.index.get_level_values(1).tolist()[0] == '04-01'
    assert percentiles_.index.get_level_values(0).tolist()[0] == 1
    # all percentiles should be NaN because demo dataset is 1 year only
    assert percentiles_.isna().all().all()
    # make a bigger dummy dataset so values are not NaN
    df = pd.DataFrame({
        'data': np.random.random(
            len(pd.date_range('2000-01-01', '2020-12-31'))),
        'date': pd.date_range('2000-01-01', '2020-12-31')})
    # test the function and check that the percentiles are not NaN
    percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day(
        df, 'data', date_column_name='date')
    assert not percentiles_.isna().all().all()
    # test a longer dummy set that exceeds 1 year
    df = pd.DataFrame({
        'data': np.random.random(
            len(pd.date_range('2000-01-01', '2001-12-31'))),
        'date': pd.date_range('2000-01-01', '2001-12-31')})
    # test the function
    percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day(
        df, 'data', date_column_name='date', year_type='water')
    assert percentiles_.shape == (365, 8)
    assert percentiles_.columns.tolist() == [0, 5, 10, 25, 75, 90, 95, 100]
    assert percentiles_.index.get_level_values(1).tolist()[0] == '10-01'
    assert percentiles_.index.get_level_values(0).tolist()[0] == 1
    assert percentiles_.index.get_level_values(1).tolist()[-1] == '09-30'
    assert percentiles_.index.get_level_values(0).tolist()[-1] == 365
    # test a longer dummy set that exceeds 1 year
    df = pd.DataFrame({
        'data': np.random.random(
            len(pd.date_range('2001-01-01', '2002-12-31'))),
        'date': pd.date_range('2001-01-01', '2002-12-31')})
    # test the function
    percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day(
        df, 'data', date_column_name='date', year_type='water')
    assert percentiles_.shape == (365, 8)
    assert percentiles_.columns.tolist() == [0, 5, 10, 25, 75, 90, 95, 100]
    assert percentiles_.index.get_level_values(1).tolist()[0] == '10-01'
    assert percentiles_.index.get_level_values(0).tolist()[0] == 1
    assert percentiles_.index.get_level_values(1).tolist()[-1] == '09-30'
    assert percentiles_.index.get_level_values(0).tolist()[-1] == 365



"""Unit tests for the exceedance module."""
import pytest
import numpy as np
import pandas as pd
from hyswap import exceedance


def test_calculate_exceedance_probability_from_distribution():
    """Test the calculate_exceedance_probability_from_distribution function."""
    # Normal and lognormal distribution tests
    x = 1
    mu = 1
    sigma = 0.25
    prob = exceedance.calculate_exceedance_probability_from_distribution(
        x, 'lognormal', mu, sigma)
    assert prob == 0.6132049428659028
    prob = exceedance.calculate_exceedance_probability_from_distribution(
        x, 'normal', mu, sigma)
    assert prob == 0.5
    # Weibull and exponential distribution tests
    k = 1
    lambda_ = 1
    prob = exceedance.calculate_exceedance_probability_from_distribution(
        x, 'weibull', k, lambda_)
    assert prob == 0.36787944117144233
    prob = exceedance.calculate_exceedance_probability_from_distribution(
        x, 'exponential', lambda_)
    assert prob == 1.0
    # Test invalid values
    with pytest.raises(TypeError):
        exceedance.calculate_exceedance_probability_from_distribution(
            x, 'exponential', mu, sigma, k, lambda_)
    with pytest.raises(ValueError):
        exceedance.calculate_exceedance_probability_from_distribution(
            x, "invalid", mu, sigma)
    with pytest.raises(TypeError):
        exceedance.calculate_exceedance_probability_from_distribution(
            x, 'exponential', mu, sigma, k=1, lambda_=1)
    with pytest.raises(TypeError):
        exceedance.calculate_exceedance_probability_from_distribution(
            "invalid", 'exponential', mu, sigma)
    with pytest.raises(TypeError):
        exceedance.calculate_exceedance_probability_from_distribution(
            x, 5, mu, sigma)


def test_calculate_exceedance_probability_from_values():
    """Test the calculate_exceedance_probability_from_values function."""
    # Testing low value of 1
    x = 1
    values_to_compare = np.array([1, 2, 3, 4])
    prob = exceedance.calculate_exceedance_probability_from_values(
        x, values_to_compare)
    assert prob == 1.0
    values_to_compare = [1, 1, 1, 1]  # use a list
    prob = exceedance.calculate_exceedance_probability_from_values(
        x, values_to_compare)
    assert prob == 1.0
    values_to_compare = pd.Series([2, 2, 2, 2])  # use a pandas series
    prob = exceedance.calculate_exceedance_probability_from_values(
        x, values_to_compare)
    assert prob == 1.0
    # Testing high value of 5
    x = 5
    values_to_compare = np.array([1, 2, 3, 4])
    prob = exceedance.calculate_exceedance_probability_from_values(
        x, values_to_compare)
    assert prob == 0.0
    values_to_compare = np.array([1, 1, 1, 1])
    prob = exceedance.calculate_exceedance_probability_from_values(
        x, values_to_compare)
    assert prob == 0.0
    # Testing middle value of 3
    x = 3
    values_to_compare = np.array([1, 2, 3, 4])
    prob = exceedance.calculate_exceedance_probability_from_values(
        x, values_to_compare)
    assert prob == 0.5
    values_to_compare = np.array([1, 1, 1, 1])
    prob = exceedance.calculate_exceedance_probability_from_values(
        x, values_to_compare)
    assert prob == 0.0
    # Testing invalid values
    with pytest.raises(TypeError):
        exceedance.calculate_exceedance_probability_from_values(
            x, "invalid")
    with pytest.raises(TypeError):
        exceedance.calculate_exceedance_probability_from_values(
            "invalid", values_to_compare)


def test_calculate_exceedance_probability_from_distribution_multiple():
    """Test calculate_exceedance_probability_from_distribution_multiple."""
    # Single values should return the same as the single value function
    x = 1
    mu = 1
    sigma = 0.25
    prob = exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
        [x], 'lognormal', mu, sigma)
    prob_single = exceedance.calculate_exceedance_probability_from_distribution(  # noqa: E501
        x, 'lognormal', mu, sigma)
    assert prob == prob_single
    prob = exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
        [x], 'normal', mu, sigma)
    prob_single = exceedance.calculate_exceedance_probability_from_distribution(  # noqa: E501
        x, 'normal', mu, sigma)
    assert prob == prob_single
    k = 1
    lambda_ = 1
    prob = exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
        [x], 'weibull', k, lambda_)
    prob_single = exceedance.calculate_exceedance_probability_from_distribution(  # noqa: E501
        x, 'weibull', k, lambda_)
    assert prob == prob_single
    prob = exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
        [x], 'exponential', lambda_)
    prob_single = exceedance.calculate_exceedance_probability_from_distribution(  # noqa: E501
        x, 'exponential', lambda_)
    assert prob == prob_single
    # Test multiple values for normal and lognormal distributions
    values = np.array([1.0, 1.25, 1.5, 1.75, 2.0])
    mu = 1
    sigma = 0.25
    prob = exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
        values, 'lognormal', mu, sigma)
    assert np.allclose(prob, np.array([0.61320494, 0.5, 0.41171189,
                                       0.34256783, 0.28787077]))
    prob = exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
        values, 'normal', mu, sigma)
    assert np.allclose(prob, np.array([0.5, 0.158655254, 0.0227501319,
                                       0.00134989803, 0.0000316712418]))
    # Test weibull and exponential distributions
    values = np.array([1, 2, 3, 4])
    k = 1
    lambda_ = 1
    prob = exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
        values, 'weibull', k, lambda_)
    assert np.allclose(prob, np.array([0.36787944, 0.13533528,
                                       0.04978707, 0.01831564]))
    prob = exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
        values, 'exponential', lambda_)
    assert np.allclose(prob, np.array([1., 0.36787944,
                                       0.13533528, 0.04978707]))
    # Test invalid values
    with pytest.raises(TypeError):
        exceedance.calculate_exceedance_probability_from_distribution_multiple(
            values, 'exponential', mu, sigma, k, lambda_)
    with pytest.raises(ValueError):
        exceedance.calculate_exceedance_probability_from_distribution_multiple(
            values, "invalid", mu, sigma)


def test_calculate_exceedance_probability_from_values_multiple():
    """Test calculate_exceedance_probability_from_values_multiple function."""
    values = np.array([1, 2, 3, 4])
    values_to_compare = np.array([1, 2, 3, 4])
    prob = exceedance.calculate_exceedance_probability_from_values_multiple(
        values, values_to_compare)
    assert np.allclose(prob, np.array([1.0, 0.75, 0.5, 0.25]))
    values_to_compare = np.array([1, 1, 1, 1])
    prob = exceedance.calculate_exceedance_probability_from_values_multiple(
        values, values_to_compare)
    assert np.allclose(prob, np.array([1.0, 0.0, 0.0, 0.0]))
    values_to_compare = np.array([2, 2, 2, 2])
    prob = exceedance.calculate_exceedance_probability_from_values_multiple(
        values, values_to_compare)
    assert np.allclose(prob, np.array([1.0, 1.0, 0.0, 0.0]))
    values_to_compare = np.array([1, 2, 3, 4, 1, 2, 3, 4])
    prob = exceedance.calculate_exceedance_probability_from_values_multiple(
        values, values_to_compare)
    assert np.allclose(prob, np.array([1.0, 0.75, 0.5, 0.25]))
    values_to_compare = np.array([1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4])
    prob = exceedance.calculate_exceedance_probability_from_values_multiple(
        values, values_to_compare)
    assert np.allclose(prob, np.array([1.0, 0.75, 0.5, 0.25]))



"""Exceedance probability calculations."""

import numpy as np
import pandas as pd
from scipy import stats


def calculate_exceedance_probability_from_distribution(x, dist,
                                                       *args, **kwargs):
    """
    Calculate the exceedance probability of a value relative to a distribution.

    Parameters
    ----------
    x : float
        The value for which to calculate the exceedance probability.
    dist : str
        The distribution to use. Must be one of 'lognormal', 'normal',
        'weibull', or 'exponential'.
    *args
        Positional arguments to pass to the distribution, which is one of
        `stats.lognorm.sf`, `stats.norm.sf`, `stats.exponweib.sf` or
        `stats.expon.sf`, refer to the `scipy.stats` documentation for more
        information about these arguments.
    **kwargs
        Keyword arguments to pass to the distribution, which is one of
        `stats.lognorm.sf`, `stats.norm.sf`, `stats.exponweib.sf` or
        `stats.expon.sf`, refer to the `scipy.stats` documentation for more
        information about these arguments.

    Returns
    -------
    float
        The exceedance probability.

    Examples
    --------
    Calculating the exceedance probability of a value of 1 from a lognormal
    distribution with a mean of 1 and a standard deviation of 0.25.

    .. doctest::

        >>> exceedance.calculate_exceedance_probability_from_distribution(
        ...     1, 'lognormal', 1, 0.25)
        0.6132049428659028

    Calculating the exceedance probability of a value of 1 from a normal
    distribution with a mean of 1 and a standard deviation of 0.25.

    .. doctest::

        >>> exceedance.calculate_exceedance_probability_from_distribution(
        ...     1, 'normal', 1, 0.25)
        0.5
    """
    # type check
    if not isinstance(x, (int, float, np.int64, np.float64, np.float32,
                          np.float16, np.int32, np.int16, np.int8)):
        raise TypeError("x must be a float (or integer).")
    if not isinstance(dist, str):
        raise TypeError("dist must be a string.")
    # do calculation
    if dist == 'lognormal':
        return stats.lognorm.sf(x, *args, **kwargs)
    elif dist == 'normal':
        return stats.norm.sf(x, *args, **kwargs)
    elif dist == 'weibull':
        return stats.exponweib.sf(x, *args, **kwargs)
    elif dist == 'exponential':
        return stats.expon.sf(x, *args, **kwargs)
    else:
        raise ValueError("dist must be one of 'lognormal', 'normal'," +
                         "'weibull', or 'exponential'.")


def calculate_exceedance_probability_from_values(x, values_to_compare):
    """
    Calculate the exceedance probability of a value compared to several values.

    This function specifically counts the number of values from the input
    `values_to_compare` that are *greater than or equal to* the input value
    `x`. The choice of greater than or equal, as opposed to solely greater than
    is intentional and follows established USGS practices [1]_.

    .. [1] Searcy, J. K. "Flow-duration curves: Water Supply Paper 1542-A."
           US Geological Survey, Reston, VA (1959).

    Parameters
    ----------
    x : float
        The value for which to calculate the exceedance probability.
    values_to_compare : array-like
        The values to use to calculate the exceedance probability.

    Returns
    -------
    float
        The exceedance probability.

    Examples
    --------
    Calculating the exceedance probability of a value of 1 from a set of values
    of 1, 2, 3, and 4.

    .. doctest::

        >>> exceedance.calculate_exceedance_probability_from_values(
        ...     1, [1, 2, 3, 4])
        1.0

    Calculating the exceedance probability of a value of 5 from a set of values
    of 1, 2, 3, and 4.

    .. doctest::

        >>> exceedance.calculate_exceedance_probability_from_values(
        ...     5, [1, 2, 3, 4])
        0.0
    """
    # some type conversions to get to a numpy array
    if isinstance(values_to_compare, list):
        values_to_compare = np.array(values_to_compare)
    elif isinstance(values_to_compare, pd.Series):
        values_to_compare = values_to_compare.values
    # raise error if not a numpy array
    if not isinstance(values_to_compare, np.ndarray):
        raise TypeError("values_to_compare must be a numpy array, list, " +
                        "or pandas Series.")
    # calculate the exceedance probability
    return np.sum(values_to_compare >= x) / len(values_to_compare)


def calculate_exceedance_probability_from_distribution_multiple(values, dist,
                                                                *args,
                                                                **kwargs):
    """
    Calculate the exceedance probability of multiple values vs a distribution.

    Parameters
    ----------
    values : array-like
        The values for which to calculate the exceedance probability.
    dist : str
        The distribution to use. Must be one of 'lognormal', 'normal',
        'weibull', or 'exponential'.
    *args
        Positional arguments to pass to the distribution, which is one of
        `stats.lognorm.sf`, `stats.norm.sf`, `stats.exponweib.sf` or
        `stats.expon.sf`, refer to the `scipy.stats` documentation for more
        information about these arguments.
    **kwargs
        Keyword arguments to pass to the distribution, which is one of
        `stats.lognorm.sf`, `stats.norm.sf`, `stats.exponweib.sf` or
        `stats.expon.sf`, refer to the `scipy.stats` documentation for more
        information about these arguments.

    Returns
    -------
    array-like
        The exceedance probabilities.

    Examples
    --------
    Calculating the exceedance probability of a set of values of 1, 1.25 and
    1.5 from a lognormal distribution with a mean of 1 and a standard
    deviation of 0.25.

    .. doctest::

        >>> exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
        ...     [1, 1.25, 1.5], 'lognormal', 1, 0.25)
        array([0.61320494, 0.5       , 0.41171189])

    Calculating the exceedance probability of a set of values of 1, 2, 3, and 4
    from a normal distribution with a mean of 1 and a standard deviation of
    0.25.

    .. doctest::

        >>> exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
        ...     [1, 2, 3, 4], 'normal', 1, 0.25)
        array([5.00000000e-01, 3.16712418e-05, 6.22096057e-16, 1.77648211e-33])
    """
    return np.array([calculate_exceedance_probability_from_distribution(
        x, dist, *args, **kwargs) for x in values])


def calculate_exceedance_probability_from_values_multiple(values,
                                                          values_to_compare):
    """
    Calculate the exceedance probability of multiple values vs a set of values.

    Parameters
    ----------
    values : array-like
        The values for which to calculate the exceedance probability.
    values_to_compare : array-like
        The values to use to calculate the exceedance probability.

    Returns
    -------
    array-like
        The exceedance probabilities.

    Examples
    --------
    Calculating the exceedance probability of a set of values of 1, 1.25 and
    1.5 from a set of values of 1, 2, 3, and 4.

    .. doctest::

        >>> exceedance.calculate_exceedance_probability_from_values_multiple(
        ...     [1, 1.25, 2.5], [1, 2, 3, 4])
        array([1.  , 0.75, 0.5 ])

    Fetch some data from NWIS and calculate the exceedance probability for a
    set of 5 values spaced evenly between the minimum and maximum values.

    .. doctest::
        :skipif: True  # skips this block of code as it broke CI pipeline

        >>> df, _ = dataretrieval.nwis.get_gwlevels(site='434400121275801',
        ...                                         start='2000-01-01',
        ...                                         end='2020-01-01')
        >>> values = np.linspace(df['lev_va'].min(),
        ...                      df['lev_va'].max(), 5)
        >>> exceedance.calculate_exceedance_probability_from_values_multiple(
        ...     values, df['lev_va'])
        array([1.        , 0.96363636, 0.83636364, 0.47272727, 0.01818182])
    """
    return np.array([calculate_exceedance_probability_from_values(
        x, values_to_compare) for x in values])



from importlib.metadata import version
from importlib.metadata import PackageNotFoundError
from hyswap.utils import *  # noqa
from hyswap.exceedance import *  # noqa
from hyswap.rasterhydrograph import *  # noqa
from hyswap.percentiles import *  # noqa
from hyswap.cumulative import *  # noqa
from hyswap.plots import *  # noqa

try:
    __version__ = version('hyswap')
except PackageNotFoundError:
    __version__ = "version-unknown"



"""Tests for the cumulative functions."""

import numpy as np
import pandas as pd
from hyswap import cumulative


def test_tidy_cumulative_dataframe():
    """Test the _tidy_cumulative_dataframe function."""
    # make a dataframe
    cdf = pd.DataFrame(
        index=np.arange(2000, 2003),
        columns=np.arange(1, 367),
        data=np.arange(1, 1099).reshape(3, 366))
    # test the function
    cdf = cumulative._tidy_cumulative_dataframe(cdf, 'calendar')
    assert cdf.shape == (1098, 3)
    assert cdf.columns.tolist() == ['index_year', 'index_doy', 'cumulative']
    assert cdf.index.year.unique().tolist() == [2000, 2001, 2002, 2003]
    assert cdf['index_doy'].tolist() == list(range(1, 367)) * 3
    assert cdf['cumulative'].tolist() == list(range(1, 1099))


def test_calculate_daily_cumulative_values():
    """Test the calculate_daily_cumulative_values function."""
    # make a dataframe
    df = pd.DataFrame({
        'data': np.arange(1, 366),
        'date': pd.date_range('2019-01-01', '2019-12-31')})
    # test the function
    cdf = cumulative.calculate_daily_cumulative_values(
        df, 'data', date_column_name='date')
    assert cdf.shape == (365, 3)
    assert cdf.columns.tolist() == ['index_year', 'index_doy', 'cumulative']
    assert cdf.index.year.unique().tolist() == [2019]
    assert cdf['index_doy'].tolist() == list(range(1, 366))
    assert cdf['cumulative'].tolist() == list(np.cumsum(range(1, 366)))
    # test the function with no date column and dates in the index
    df = df.set_index('date')
    cdf = cumulative.calculate_daily_cumulative_values(df, 'data')
    assert cdf.shape == (365, 3)
    assert cdf.columns.tolist() == ['index_year', 'index_doy', 'cumulative']
    assert cdf.index.year.unique().tolist() == [2019]
    assert cdf['index_doy'].tolist() == list(range(1, 366))
    assert cdf['cumulative'].tolist() == list(np.cumsum(range(1, 366)))
    # try for a water year
    df = pd.DataFrame({
        'data': np.ones(len(pd.date_range('2016-01-01', '2019-12-31'))),
        'date': pd.date_range('2016-01-01', '2019-12-31')})
    cdf = cumulative.calculate_daily_cumulative_values(
        df, 'data', date_column_name='date', year_type='water')
    assert cdf.shape == (365*3, 3)
    assert cdf.columns.tolist() == ['index_year', 'index_doy', 'cumulative']
    assert cdf.index.year.unique().tolist() == [2017, 2018, 2019, 2020]
    assert cdf['index_doy'].tolist()[0] == 1
    d_list = cdf['index_doy'].unique().tolist()
    d_list.sort()
    assert d_list == list(range(1, 366))
    assert cdf['cumulative'].tolist()[:365] == list(np.cumsum(np.ones(365)))
    # try for a climate year
    cdf = cumulative.calculate_daily_cumulative_values(
        df, 'data', date_column_name='date', year_type='climate')
    assert cdf.shape == (365*3, 3)
    assert cdf.columns.tolist() == ['index_year', 'index_doy', 'cumulative']
    assert cdf.index.year.unique().tolist() == [2017, 2018, 2019, 2020]
    assert cdf['index_doy'].tolist()[0] == 1
    d_list = cdf['index_doy'].unique().tolist()
    d_list.sort()
    assert d_list == list(range(1, 366))
    assert cdf['cumulative'].tolist()[:365] == list(np.cumsum(np.ones(365)))



"""Functions to make plots."""
import calendar
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from hyswap.percentiles import calculate_variable_percentile_thresholds_by_day


def plot_flow_duration_curve(
        values, exceedance_probabilities, ax=None, title='Flow Duration Curve',
        xlab='Exceedance Probability\n' +
        '(Percent of Time Indicated Discharge was Equaled or Exceeded)',
        ylab='Discharge, ft$^3$/s', grid=True, **kwargs):
    """
    Make flow duration curve plot.

    Parameters
    ----------
    values : array-like
        Values to plot along y-axis.
    exceedance_probabilities : array-like
        Exceedance probabilities for each value, likely calculated from
        a function like :obj:`hyswap.exceedance.calculate_exceedance_probability_from_values_multiple`.
    ax : matplotlib.axes.Axes, optional
        Axes to plot on. If not provided, a new figure and axes will be
        created.
    title : str, optional
        Title for the plot. If not provided, the default title will be
        'Flow Duration Curve'.
    xlab : str, optional
        Label for the x-axis. If not provided, a default label will be used.
    ylab : str, optional
        Label for the y-axis. If not provided, a default label will be used.
    grid : bool, optional
        Whether to show grid lines on the plot. Default is True.
    **kwargs
        Keyword arguments passed to :meth:`matplotlib.axes.Axes.plot`.

    Returns
    -------
    matplotlib.axes.Axes
        Axes object containing the plot.

    Examples
    --------
    Fetch some data from NWIS, calculate the exceedance probabilities and then
    make the flow duration curve.

    .. plot::
        :include-source:

        >>> df, _ = dataretrieval.nwis.get_dv(site='06892350',
        ...                                   parameterCd='00060',
        ...                                   start='1776-07-04',
        ...                                   end='2020-01-01')
        >>> values = np.linspace(df['00060_Mean'].min(),
        ...                      df['00060_Mean'].max(), 10000)
        >>> exceedance_probabilities = hyswap.exceedance.calculate_exceedance_probability_from_values_multiple(  # noqa
        ...     values, df['00060_Mean'])
        >>> ax = hyswap.plots.plot_flow_duration_curve(
        ...     values, exceedance_probabilities,
        ...     title='Flow Duration Curve for USGS Site 06892350')
        >>> plt.tight_layout()
        >>> plt.show()
    """
    # Create axes if not provided
    if ax is None:
        _, ax = plt.subplots()
    # do plotting
    ax.plot(exceedance_probabilities*100, values, **kwargs)
    ax.set_xlabel(xlab)
    ax.set_ylabel(ylab)
    ax.set_title(title)
    # set log scales for x axis
    ax.set_yscale('log')
    # set limits for axes
    ax.set_xlim(0.1, 99.9)
    # set ticks for axes
    # always use same ticks for x-axis
    ax.set_xticks([0.1, 5, 10, 25, 50, 75, 90, 95, 99.9])
    ax.set_xticklabels([
        '0.1', '5', '10', '25', '50', '75', '90', '95', '99.9'])
    # get y-axis ticks and convert to comma-separated strings
    yticks = ax.get_yticks()
    yticklabels = [f'{int(y):,}' for y in yticks]
    ax.set_yticks(yticks, labels=yticklabels)
    # add grid lines
    if grid:
        ax.grid(which='both', axis='both', alpha=0.5)
    # return the axes
    return ax


def plot_raster_hydrograph(df_formatted, ax=None,
                           title='Streamflow Raster Hydrograph',
                           xlab='Month', ylab='Year',
                           cbarlab='Streamflow, cubic feet per second',
                           **kwargs):
    """Make raster hydrograph plot.

    Parameters
    ----------
    df_formatted : pandas.DataFrame
        Formatted dataframe containing the raster hydrograph data.
    ax : matplotlib.axes.Axes, optional
        Axes to plot on. If not provided, a new figure and axes will be
        created.
    title : str, optional
        Title for the plot. If not provided, the default title will be
        'Streamflow Raster Hydrograph'.
    xlab : str, optional
        Label for the x-axis. If not provided, the default label will be
        'Month'.
    ylab : str, optional
        Label for the y-axis. If not provided, the default label will be
        'Year'.
    cbarlab : str, optional
        Label for the colorbar. If not provided, the default label will be
        'Streamflow, cubic feet per second'.
    **kwargs
        Keyword arguments passed to :meth:`matplotlib.axes.Axes.imshow`.

    Returns
    -------
    matplotlib.axes.Axes
        Axes object containing the plot.

    Examples
    --------
    Fetch some data from NWIS, format it for a raster hydrograph plot and then
    make the raster hydrograph plot.

    .. plot::
        :include-source:

        >>> df, _ = dataretrieval.nwis.get_dv(site='09380000',
        ...                                   parameterCd='00060',
        ...                                   start='1960-01-01',
        ...                                   end='1970-12-31')
        >>> df_rh = hyswap.rasterhydrograph.format_data(df, '00060_Mean')
        >>> fig, ax = plt.subplots(figsize=(6, 6))
        >>> ax = hyswap.plots.plot_raster_hydrograph(
        ...     df_rh, ax=ax, title='Raster Hydrograph for USGS Site 09380000')
        >>> plt.tight_layout()
        >>> plt.show()
    """
    # Create axes if not provided
    if ax is None:
        _, ax = plt.subplots()
    # define min/max values
    min_10 = np.nanmax(
        [np.floor(np.log10(np.nanmin(df_formatted.to_numpy()))), 0]
    )
    max_10 = np.ceil(np.log10(np.nanmax(df_formatted.to_numpy())))
    # pop some kwargs
    cmap = kwargs.pop('cmap', 'YlGnBu')
    aspect = kwargs.pop('aspect', 'auto')
    interpolation = kwargs.pop('interpolation', 'none')
    vmin = kwargs.pop('vmin', int(10**min_10))
    vmax = kwargs.pop('vmax', int(10**max_10))
    norm = kwargs.pop('norm', matplotlib.colors.LogNorm(vmin=vmin, vmax=vmax))
    # do plotting
    img = ax.imshow(df_formatted, aspect=aspect, cmap=cmap,
                    interpolation=interpolation, norm=norm, **kwargs)
    # set labels
    ax.set_xlabel(xlab)
    ax.set_ylabel(ylab)
    ax.set_title(title)
    # add colorbar
    cbar = plt.colorbar(img, ax=ax)
    # set colorbar ticks
    cticks = cbar.ax.get_yticks()
    cbar.ax.set_yticks(cticks[1:-1],
                       labels=[f'{int(v):,}' for v in cticks[1:-1]])
    # set colorbar label
    cbar.set_label(cbarlab)
    # cbar height to be same as axes
    cbar.ax.set_aspect('auto')
    # set yticks
    ax.set_yticks(np.arange(-0.5, len(df_formatted.index)), [], minor=True)
    ax.set_yticks(np.arange(len(df_formatted.index)), df_formatted.index)
    # convert yticklabels to 4 digit years
    yticklabels = ax.get_yticklabels()
    yticklabels = [f'{y.get_text()[:4]}' for y in yticklabels]
    ax.set_yticklabels(yticklabels)
    # set xticks at start/end of each month
    xvals = df_formatted.columns.values
    months = [int(i.split('-')[1]) for i in xvals]
    month_transitions = np.where(np.diff(months) != 0)[0]
    ax.set_xticks([0] + list(month_transitions),
                  labels=[], minor=False)
    # set xticklabels to be month name at middle of each month
    unique_months = []
    [unique_months.append(x) for x in months if x not in unique_months]
    month_names = [calendar.month_abbr[i] for i in unique_months]
    month_names = [f'{m}' for m in month_names]
    days = [int(i.split('-')[2]) for i in xvals]
    midway_pts = np.where(np.array(days) == 15)[0]
    ax.set_xticks(midway_pts, labels=month_names, minor=True)
    # make minor ticks invisible
    ax.tick_params(which='minor', length=0)
    # return axes
    return ax


def plot_duration_hydrograph(percentiles_by_day, df, data_col, doy_col,
                             pct_list=[0, 5, 10, 25, 75, 90, 95, 100],
                             data_label=None, ax=None,
                             title="Duration Hydrograph",
                             ylab="Discharge (cfs)", xlab="Month",
                             colors=None, **kwargs):
    """Make duration hydrograph plot.

    Parameters
    ----------
    percentiles_by_day : pandas.DataFrame
        Dataframe containing the percentiles by day.
    df : pandas.DataFrame
        Dataframe containing the data to plot.
    data_col : str
        Column name of the data to plot.
    doy_col : str
        Column name of the day of year.
    pct_list : list, optional
        List of integers corresponding to the percentile values to be
        plotted. Defaults to 0, 5, 10, 25, 75, 90, 95, 100.
    data_label : str, optional
        Label for the data to plot. If not provided, a default label will
        be used.
    ax : matplotlib.axes.Axes, optional
        Axes to plot on. If not provided, a new figure and axes will be
        created.
    title : str, optional
        Title for the plot. If not provided, the default title will be
        'Duration Hydrograph'.
    ylab : str, optional
        Label for the y-axis. If not provided, the default label will be
        'Discharge (cfs)'.
    xlab : str, optional
        Label for the x-axis. If not provided, the default label will be
        'Month'.
    colors : list, optional
        List of colors to use for the lines. If not provided, a default
        list of colors will be used.
    **kwargs
        Keyword arguments passed to :meth:`matplotlib.axes.Axes.fill_between`.

    Returns
    -------
    matplotlib.axes.Axes
        Axes object containing the plot.

    Examples
    --------
    Fetch some data from NWIS and make a streamflow duration hydrograph plot.

    .. plot::
        :include-source:

        >>> df, _ = dataretrieval.nwis.get_dv(site='06892350',
        ...                                   parameterCd='00060',
        ...                                   start='1900-01-01',
        ...                                   end='2022-12-31')
        >>> pct_by_day = hyswap.percentiles.calculate_variable_percentile_thresholds_by_day(  # noqa: E501
        ...     df, '00060_Mean')
        >>> df_2022 = df[df.index.year == 2022]
        >>> df_2022['doy'] = df_2022.index.dayofyear
        >>> df_2022 = df_2022.sort_values(by='doy')
        >>> fig, ax = plt.subplots(figsize=(12, 6))
        >>> ax = hyswap.plots.plot_duration_hydrograph(
        ...     pct_by_day, df_2022, '00060_Mean', 'doy',
        ...     data_label='2022 Daily Mean Discharge',
        ...     ax=ax, title='Duration Hydrograph for USGS Site 06892350')
        >>> plt.tight_layout()
        >>> plt.show()
    """
    # Create axes if not provided
    if ax is None:
        _, ax = plt.subplots()
    # pop some kwargs
    alpha = kwargs.pop('alpha', 0.5)
    zorder = kwargs.pop('zorder', -20)
    if data_label is None:
        label = df[data_col].name
    else:
        label = data_label
    # get colors
    if colors is None:
        colors = ["#e37676", "#e8c285", "#dbf595", "#a1cc9f",
                  "#7bdbd2", "#7587bf", "#ad63ba"]
    # plot the latest data -1 to 0-index day of year
    ax.plot(df[doy_col]-1, df[data_col], color='k', zorder=10, label=label)
    # plot the historic percentiles filling between each pair
    pct_list.sort()  # sort the list in ascending order
    for i in range(1, len(pct_list)):
        ax.fill_between(
            percentiles_by_day.index.get_level_values(1),
            list(percentiles_by_day[pct_list[i - 1]].values),
            list(percentiles_by_day[pct_list[i]].values),
            color=colors[i - 1],
            alpha=alpha,
            linewidth=0,
            label="{}th - {}th Percentile".format(
                pct_list[i - 1], pct_list[i]),
            zorder=zorder, **kwargs
        )
    # set labels
    ax.set_xlabel(xlab)
    ax.set_xlim(0, 365)
    # major xticks at first/end of each month
    months = [int(m.split('-')[0]) for m in percentiles_by_day.index.get_level_values(1).to_list()]  # noqa: E501
    month_switch = np.where(np.diff(months) != 0)[0]
    ax.set_xticks([0] + list(month_switch + 1) + [365], labels=[], minor=False)
    # minor xticks at 15th of each month
    unique_months = []
    [unique_months.append(x) for x in months if x not in unique_months]
    month_names = [calendar.month_abbr[i] for i in unique_months]
    month_names = [f'{m}' for m in month_names]
    days = [int(m.split('-')[1]) for m in percentiles_by_day.index.get_level_values(1).to_list()]  # noqa: E501
    mid_days = np.where(np.array(days) == 15)[0]
    ax.set_xticks(list(mid_days + 1), labels=month_names, minor=True)
    # make minor ticks invisible
    ax.tick_params(axis='x', which='minor', length=0)
    # other labels
    ax.set_ylabel(ylab)
    ax.set_yscale("log")
    ax.set_title(title)
    # get y-axis ticks and convert to comma-separated strings
    yticks = ax.get_yticks()
    yticklabels = [f'{int(y):,}' for y in yticks]
    ax.set_yticks(yticks[1:-1], labels=yticklabels[1:-1])
    # two column legend
    ax.legend(loc="best", ncol=2)
    # return axes
    return ax


def plot_cumulative_hydrograph(cumulative_percentiles, target_year,
                               year_type='calendar',
                               envelope_pct=[25, 75],
                               max_pct=False, min_pct=False,
                               ax=None,
                               title="Cumulative Discharge",
                               ylab="Cumulative Discharge (cfs)",
                               xlab="Month", **kwargs):
    """Make cumulative hydrograph plot.

    Parameters
    ----------
    cumulative_percentiles : pandas.DataFrame
        Dataframe containing the cumulative percentiles per year, output
        from :obj:`hyswap.cumulative.calculate_daily_cumulative_values`.
    target_year : int
        Target year to plot in black as the line.
    year_type : str, optional
        The type of year to use. Must be one of 'calendar', 'water', or
        'climate'. Default is 'calendar' which starts the year on January 1
        and ends on December 31. 'water' starts the year on October 1 and
        ends on September 30 of the following year which is the "water year".
        For example, October 1, 2010 to September 30, 2011 is "water year
        2011". 'climate' years begin on April 1 and end on March 31 of the
        following year, they are numbered by the ending year. For example,
        April 1, 2010 to March 31, 2011 is "climate year 2011".
    envelope_pct : list, optional
        List of percentiles to plot as the envelope. Default is [25, 75].
    max_pct : bool, optional
        If True, plot the maximum value as a dashed line. Default is False.
    min_pct : bool, optional
        If True, plot the minimum value as a dotted line. Default is False.
    ax : matplotlib.axes.Axes, optional
        Axes to plot on. If not provided, a new figure and axes will be
        created.
    title : str, optional
        Title for the plot. If not provided, the default title will be
        'Cumulative Discharge'.
    ylab : str, optional
        Label for the y-axis. If not provided, the default label will be
        'Cumulative Discharge (cfs)'.
    xlab : str, optional
        Label for the x-axis. If not provided, the default label will be
        'Month'.
    **kwargs
        Keyword arguments passed to :meth:`matplotlib.axes.Axes.fill_between`.

    Returns
    -------
    matplotlib.axes.Axes
        Axes object containing the plot.

    Examples
    --------
    Fetch some data from NWIS and make a cumulative hydrograph plot.

    .. plot::
        :include-source:

        >>> df, _ = dataretrieval.nwis.get_dv(site='06892350',
        ...                                   parameterCd='00060',
        ...                                   start='1900-01-01',
        ...                                   end='2021-12-31')
        >>> cdf = hyswap.cumulative.calculate_daily_cumulative_values(
        ...     df, '00060_Mean')
        >>> fig, ax = plt.subplots(figsize=(8, 5))
        >>> ax = hyswap.plots.plot_cumulative_hydrograph(
        ...     cdf, 2020, ax=ax,
        ...     title='2020 Cumulative Streamflow Hydrograph, site 06892350')
        >>> plt.tight_layout()
        >>> plt.show()
    """
    # Create axes if not provided
    if ax is None:
        _, ax = plt.subplots()
    # calculations for percentiles by day
    pdf = calculate_variable_percentile_thresholds_by_day(
        cumulative_percentiles, data_column_name='cumulative',
        percentiles=[0] + envelope_pct + [100],
        year_type=year_type)
    # pop some kwargs
    alpha = kwargs.pop('alpha', 0.5)
    zorder = kwargs.pop('zorder', -20)
    color = kwargs.pop('color', 'xkcd:bright green')
    # plot 25-75 percentile envelope
    ax.fill_between(pdf.index.get_level_values(1),
                    list(pdf[envelope_pct[0]].values),
                    list(pdf[envelope_pct[1]].values),
                    color=color, alpha=alpha,
                    label=f"{envelope_pct[0]}th - {envelope_pct[1]}th " +
                          "Percentile Envelope",
                    zorder=zorder, **kwargs)
    # plot min/max if desired
    if min_pct:
        ax.plot(pdf.index.get_level_values(1), pdf[0], color='k',
                alpha=0.5, linestyle=':', label="Minimum")
    if max_pct:
        ax.plot(pdf.index.get_level_values(1), pdf[100], color='k',
                alpha=0.5, linestyle='--', label="Maximum")
    # get data from target year
    target_year_data = cumulative_percentiles.loc[
        cumulative_percentiles['index_year'] == target_year]
    # plot target year
    ax.plot(target_year_data['index_doy'], target_year_data['cumulative'],
            color='k', label=f"{target_year} Cumulative Discharge")

    # set labels
    ax.set_xlabel(xlab)
    ax.set_xlim(0, 365)
    # major xticks at first/end of each month
    months = [int(m.split('-')[0]) for m in pdf.index.get_level_values(1).to_list()]  # noqa: E501
    month_switch = np.where(np.diff(months) != 0)[0]
    ax.set_xticks([0] + list(month_switch + 1) + [365], labels=[], minor=False)
    # minor xticks at 15th of each month
    unique_months = []
    [unique_months.append(x) for x in months if x not in unique_months]
    month_names = [calendar.month_abbr[i] for i in unique_months]
    month_names = [f'{m}' for m in month_names]
    days = [int(m.split('-')[1]) for m in pdf.index.get_level_values(1).to_list()]  # noqa: E501
    mid_days = np.where(np.array(days) == 15)[0]
    ax.set_xticks(list(mid_days + 1), labels=month_names, minor=True)
    # make minor ticks invisible
    ax.tick_params(axis='x', which='minor', length=0)
    # other labels
    ax.set_ylabel(ylab)
    ax.set_title(title)
    # get y-axis ticks and convert to comma-separated strings
    yticks = ax.get_yticks()
    yticklabels = [f'{int(y):,}' for y in yticks]
    ax.set_yticks(yticks[1:-1], labels=yticklabels[1:-1])
    # two column legend
    ax.legend(loc="best")

    # return
    return ax
